local ColourScheme = { -- esquema de color
  [0] = {StyleFore = "000000", StyleBack = "EEECFF", },
  [1] = {StyleFore = "005D00", StyleBack = "B6FF8C", },
  [2] = {StyleFore = "685D1C", StyleBack = "FFE500", },
  [3] = {StyleFore = "FF0000", StyleBack = "E0ACB0", StyleBold = true,},
  [4] = {StyleFore = "000000", StyleBack = "515151", StyleItalic = true,},
}

local TodoSymbols = { -- Lista de elementos de Todo
	[0] = '..',
	[1] = '++',
	[2] = '+-',
	[3] = '--',
	[4] = '??',
}

local function SetColours(lexer, scheme)
  local function dec(s) return tonumber(s, 16) end
  if lexer then editor.Lexer = lexer end
  for i, style in pairs(scheme) do
    for prop, value  in pairs(style) do
      if (prop == "StyleFore" or prop == "StyleBack")
         and type(value) == "string" then -- convert from string
        local hex, hex, r, g, b =
          string.find(value, "^(%x%x)(%x%x)(%x%x)$")
        value = hex and (dec(r) + dec(g)*256 + dec(b)*65536) or 0
      end
      editor[prop][i] = value
    end
  end
end

function GetSchemeForSymbol(sym)
	for i, lstSym in pairs(TodoSymbols) do
		if (lstSym == sym) then
			return i
		end
	end
	return -1
end

function IsStrATodoSymbol(str)
	for i, lstSym in pairs(TodoSymbols) do
		if (lstSym == str) then
			return true
		end
	end
	return false
end

function GetCoordsBetweenTodoSymbs()
	local currPos = editor.CurrentPos

	local s = Search("back")
	editor:GotoPos(currPos)
	local e = Search("foreword")

	if (s == -666) then s=0 end
	if (e == -666) then e=editor.TextLength end


	return s, e
end

function Search(direction)
	local currPos = editor.CurrentPos
	local e = -1

	local dirOffset = 0
	local levelDownChar = ''
	local levelUpChar = ''
	local currPosLimit = 0

	if (direction == 'back') then
		dirOffset = -1
		currPosLimit = -1
		currPos = currPos - 1
	else
		dirOffset = 1
		currPosLimit = editor.TextLength
	end

	while e == -1 do
		local strToLook

		if (direction == 'back') then
			strToLook = Global.CharAt(currPos-1)..Global.CharAt(currPos)
		else
			strToLook = Global.CharAt(currPos)..Global.CharAt(currPos+1)
		end

		if ( IsStrATodoSymbol(strToLook) ) then
			if (direction == 'back') then
				e = currPos -1
			else
				e = currPos
			end
		end

		if (currPos == currPosLimit) then
			return -666
		end

		-- FOR DEBUG
		--print ('Dir:'..direction..' in char '..Global.CharAt(currPos)..' at position '..currPos)

		currPos = currPos + dirOffset
	end

	return e
end

function GetLineTodoSym(currLine)
	if (currLine == nil) then
		return nil
	end

	return string.sub(currLine,1,2)
end

function EnableTodo()

	local function Colourise()
		local currPos = editor.CurrentPos

		SetColours(SCLEX_CONTAINER, ColourScheme)

		local s, e = GetCoordsBetweenTodoSymbs()
		editor:SetSel(s, e)

		local todoSym = GetLineTodoSym(editor:GetSelText())

		editor:StartStyling(s, 31)
		if (e ~= 0) then e = e-1 end
		editor:SetStyling(e-s, GetSchemeForSymbol(todoSym))

		editor:GotoPos(currPos)
	end

	local function StartColourise()
		local currPos = editor.CurrentPos
		for p=0, string.len(editor:GetText()) do
			editor.CurrentPos = p
			Colourise()
		end
		editor:GotoPos(currPos)

		scite_OnChar(Colourise)
	end

	scite_OnOpenSwitch(function()
		if(props['FileExt'] == 'txt') then
			StartColourise()
			return
		else
			--Removemos el evento
			scite_OnChar(Colourise,true)
		end
	end)

	StartColourise()
end
